@@ -1,4 +1,4 @@
-package App::MBUtiny; # $Id: MBUtiny.pm 39 2014-08-30 08:57:38Z abalama $
+package App::MBUtiny; # $Id: MBUtiny.pm 50 2014-09-03 12:24:39Z abalama $
 use strict;
 
 =head1 NAME
@@ -7,7 +7,7 @@ App::MBUtiny - BackUp system for Your WEBsites
 
 =head1 VERSION
 
-Version 1.02
+Version 1.03
 
 =head1 SYNOPSIS
 
@@ -37,6 +37,12 @@ Returns object. $c -- CTK object
 
 Run BackUp for all or specified names of hosts
 
+=item B<restore>
+
+    my $status = $mbu->test( [qw( ... host names ... ), $date] );
+
+Restore files for all or specified hosts by date
+
 =item B<test>
 
     my $status = $mbu->test( [qw( ... host names ... )] );
@@ -55,6 +61,12 @@ Returns CTK object
 
 Set/Get message. Method returns informational message of MBUtiny
 
+=item B<show>
+
+    print $mbu->show;
+
+Set/Get message for user. Method returns informational message for user of MBUtiny
+
 =back
 
 =head1 HISTORY
@@ -102,7 +114,7 @@ See C<LICENSE> file
 =cut
 
 use vars qw/ $VERSION /;
-$VERSION = '1.02';
+$VERSION = '1.03';
 
 use CTK::Util;
 use CTK::ConfGenUtil;
@@ -122,6 +134,7 @@ use App::MBUtiny::CollectorAgent;
 use constant {
     OBJECTS_DIR => 'files',
     EXCLUDE_DIR => 'excludes',
+    RESTORE_DIR => 'restore',
 };
 
 sub new {
@@ -131,16 +144,20 @@ sub new {
     
     my $objdir = catdir($c->datadir,OBJECTS_DIR);
     my $excdir = catdir($c->datadir,EXCLUDE_DIR);
+    my $rstdir = catdir($c->datadir,RESTORE_DIR);
     preparedir({
             objdir => $objdir,
             excdir => $excdir,
+            rstdir => $rstdir,
         });
     
     my $self = bless { 
             c   => $c,
             msg => '',
+            show   => '',
             objdir => $objdir,
             excdir => $excdir,
+            rstdir => $rstdir,
         }, $class;
     
     #$c->log_debug("new say: Blah-Blah-Blah");
@@ -154,6 +171,12 @@ sub msg {
     $self->{msg} = $s if defined $s;
     return $self->{msg};
 }
+sub show { 
+    my $self = shift;
+    my $s = shift;
+    $self->{show} = $s if defined $s;
+    return $self->{show};
+}
 sub backup {
     my $self = shift;
     my $args = array(shift);
@@ -762,7 +785,7 @@ sub backup {
                 $ma{'-message'} .= "\n---\n"
                                  . sprintf("Generated by    : MBUtiny %s\n", $VERSION)
                                  . sprintf("Date generation : %s\n", localtime2date_time())
-                                 . sprintf("MBUTiny Id      : %s\n", '$Id: MBUtiny.pm 39 2014-08-30 08:57:38Z abalama $')
+                                 . sprintf("MBUTiny Id      : %s\n", '$Id: MBUtiny.pm 50 2014-09-03 12:24:39Z abalama $')
                                  . sprintf("Time Stamp      : %s\n", CTK::tms())
                                  . sprintf("Configuration   : %s\n", $c->cfgfile);
                 $ma{'-attach'} = _attach($ma{'-attach'}) || [];
@@ -1040,7 +1063,7 @@ sub test {
         $ma{'-message'} .= "\n---\n"
             . sprintf("Generated by    : MBUtiny %s\n", $VERSION)
             . sprintf("Date generation : %s\n", localtime2date_time())
-            . sprintf("MBUTiny Id      : %s\n", '$Id: MBUtiny.pm 39 2014-08-30 08:57:38Z abalama $')
+            . sprintf("MBUTiny Id      : %s\n", '$Id: MBUtiny.pm 50 2014-09-03 12:24:39Z abalama $')
             . sprintf("Time Stamp      : %s\n", CTK::tms())
             . sprintf("Configuration   : %s\n", $c->cfgfile);
         $ma{'-attach'} = _attach($ma{'-attach'}) || [];
@@ -1053,6 +1076,369 @@ sub test {
     $self->msg($rslt);
     return $gstatus;
 }
+sub restore {
+    my $self = shift;
+    my $args = array(shift);
+    my $c    = $self->c;
+    my $config = $c->config;
+    my $ret = "";
+    
+    #  
+    my @tblfields = ( # 
+            [19, 'DATE AND TIME'],
+            [32, 'PROCESS NAME'],
+            [42, 'DESCRIPTION OF PROCCESS / DATA OF PROCCESS'],
+            [8,  'STATUS'],
+        );
+        
+    #   
+    my $arcdef = $config->{arc};
+    croak "Error! Undefined <arc> section." unless $arcdef;
+    
+    #  
+    my $tdate = pop @$args;
+    my ( $_y, $_m, $_d ) = (localtime( time ))[5,4,3];
+    my @ymd = (($_y+1900), ($_m+1), $_d);
+    if (defined($tdate)) {
+        if ($tdate =~ /(\d{4})\D+(\d{2})\D+(\d{2})/) {
+            @ymd = ($1,$2,$3);
+        } elsif ($tdate =~ /(\d{2})\D+(\d{2})\D+(\d{4})/) {
+            @ymd = ($3,$2,$1);
+        } else {
+            push @$args, $tdate;
+        }
+    }
+    #$c->log_debug(sprintf(">>> %d/%d/%d", @ymd));
+    
+    #  
+    my @joblist = $self->get_jobs;
+    $c->log_debug("Start processing hosts");
+    foreach my $job (sort {(keys(%$a))[0] cmp (keys(%$b))[0]} @joblist) {
+        my $hostname = _name($job);
+        my $hostskip = (!@$args || grep {lc($hostname) eq lc($_)} @$args) ? 0 : 1;
+        $c->log_debug(sprintf("Loading configuration for host \"%s\"... %s", $hostname, ($hostskip ? 'SKIPPED' : 'LOADED') ));
+        next if $hostskip;
+        
+        #  
+        my $enabled  = value($job, $hostname => 'enable');
+        if ($enabled) {
+            $c->log_debug(sprintf("--> Begin processing: \"%s\"", $hostname));
+            my $tbl = Text::SimpleTable->new(@tblfields);
+            my $pfx = " " x 3;
+            my $step = '';
+            my $status = 0;
+            
+            #
+            # Step 00.  
+            #
+            $step = "Step 00."; $c->log_debug($pfx, $step, "Loading and preparing data");
+            
+            # C   (      )
+            my $target_dir = catdir($c->datadir, RESTORE_DIR, $hostname, sprintf("%04d%02d%02d",@ymd));
+            preparedir($target_dir);
+            
+            #    
+            my $arcname = value($job, $hostname => 'arcname') || 'tar';
+
+            #         
+            my $arcmask = value($job, $hostname => 'arcmask') || '[HOST]-[YEAR]-[MONTH]-[DAY].[EXT]';
+            $arcmask =~ s/\[DEFAULT\]/[HOST]-[YEAR]-[MONTH]-[DAY].[EXT]/gi;
+            my %maskfmt = (
+                    HOST  => $hostname,
+                    EXT   => value($arcdef, 'arc'=>$arcname=>'ext')  || '',
+                    TYPE  => value($arcdef, 'arc'=>$arcname=>'type') || '',
+                );
+            ($maskfmt{YEAR}, $maskfmt{MONTH}, $maskfmt{DAY}) = 
+                (sprintf("%04d",$ymd[0]), sprintf("%02d",$ymd[1]), sprintf("%02d",$ymd[2]));
+            my $find_file = dformat($arcmask,\%maskfmt);
+            $tbl->row(localtime2date_time, sprintf("%s Loading data", $step), $find_file, 'OK');
+            $c->log_debug($pfx x 2, $find_file);
+            
+            #####
+            # Step 01.        . 
+            #####
+            $step = "Step 01."; $c->log_debug($pfx, $step, "Get collector data");
+            my $collector_node = _node2anode(node($job, $hostname => 'collector'));
+            my $colls = [];
+            foreach my $coll (grep {value($_ => 'uri')}  @$collector_node) {
+                my $coll_uri = value($coll, 'uri') || '';
+                my $agent = new App::MBUtiny::CollectorAgent(
+                            uri         => $coll_uri,
+                            user        => value($coll, 'user'),
+                            password    => value($coll, 'password'),
+                            timeout     => value($coll, 'timeout'),
+                        );
+                my $coll_status = $agent->check;
+                
+                #  
+                $c->log_debug($pfx, sprintf(" - %s",$coll_uri));
+                $tbl->row(localtime2date_time, sprintf("%s Collector ready", $step), sprintf("%s\n%s",$coll_uri, unidecode($agent->error)), $coll_status ? 'OK' : 'ERROR');
+                if ($coll_status) {
+                    push @$colls, $agent
+                } else {
+                    $c->log_error($pfx x 2, sprintf("ERROR: %s",unidecode($agent->error)));
+                }
+            }
+
+            #####
+            # Step 02.   
+            #####
+            my $downloaded = 0; #        .    
+            my $meta; # -   ,   
+
+            # Step 02a.            
+            $step = "Step 02a."; $c->log_debug($pfx, $step, "Search file on the LOCAL storages");
+            my $localdir_node = array($job, $hostname => 'local/localdir') || [];
+            my $trylocal;
+            foreach my $localdir (@$localdir_node) {
+                my $locallist = getlist($localdir);
+                $trylocal = _search_file($locallist, $find_file);
+                if ($trylocal) { #  ! ! 
+                    my $file_src = catfile($localdir,$trylocal);
+                    my $file_dst = catfile($target_dir,$trylocal);
+                    $c->log_debug($pfx x 2, sprintf("Restore file %s from the LOCAL storage %s", $trylocal, $localdir));
+                    
+                    #   
+                    $c->fcopy(
+                        -in     => $localdir,
+                        -out    => $target_dir,
+                        -list   => $trylocal,
+                    );
+                    unless (-e $file_dst) {
+                        $c->log_error($pfx, sprintf(" - [ERROR] Can't copy file %s from %s to %s",$trylocal, $localdir, $target_dir));
+                        next
+                    }
+                    
+                    #       : 
+                    $meta = $self->_info($colls, host => $hostname, file => $trylocal) if value($job, $hostname => 'local/fixup');
+                    if ($meta) {
+                        my $size = -s $file_dst || 0;
+                        if (value($meta => "file_size") && value($meta => "file_size") ne $size){
+                            $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. Size got: %s; expected: %s",$file_dst, $size, uv2null(value($meta => "file_size"))));
+                            next;
+                        }
+                        my $md5 = md5sum($file_dst);
+                        if (value($meta => "file_md5") && value($meta => "file_md5") ne $md5) {
+                            $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. MD5 got: %s; expected: %s",$file_dst, $md5, uv2null(value($meta => "file_md5"))));
+                            next;
+                        }
+                        my $sha1 = sha1sum($file_dst);
+                        if (value($meta => "file_sha1") && value($meta => "file_sha1") ne $sha1) {
+                            $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. SHA1 got: %s; expected: %s",$file_dst, $sha1, uv2null(value($meta => "file_sha1"))));
+                            next;
+                        }
+                    }
+                    
+                    # OK!   
+                    $c->log_debug($pfx, sprintf(" - [ OK ] %s from %s",$trylocal, $localdir));
+                    $tbl->row(localtime2date_time, sprintf("%s File from LOCAL", $step), sprintf("%s from\n%s", $trylocal, $localdir), 'OK');
+                    $downloaded = 1;
+                    last;
+                } else {
+                    $c->log_debug($pfx, sprintf(" - [SKIP] %s", $localdir));
+                }
+            }
+            $c->log_debug($pfx x 2, "SKIPPED") unless $downloaded;
+            $tbl->row(localtime2date_time, sprintf("%s File from LOCAL", $step), '--- NONE ---' , 'SKIPPED') unless $downloaded;
+
+
+            # Step 02b.       FTP     
+            $step = "Step 02b."; $c->log_debug($pfx, $step, "Search file on the FTP storages");
+            if ($downloaded) {
+                $c->log_debug($pfx x 2, "SKIPPED");
+                $tbl->row(localtime2date_time, sprintf("%s File from FTP", $step), '--- NONE ---' , 'SKIPPED');
+            } else {
+                #     FTP 
+                my $ftp_node = _node2anode(node($job, $hostname => 'ftp'));
+                my $tryftp;
+                foreach my $ftpct (@$ftp_node) {
+                    my $ftpuri = sprintf("ftp://%s\@%s/%s", value($ftpct, 'ftpuser'), value($ftpct, 'ftphost'), value($ftpct, 'ftpdir'));
+                    my $ftph = ftp($ftpct, 'connect');
+                    if ($ftph) {
+                        $ftph->quit();
+                    } else {
+                        $c->log_error($pfx x 2, sprintf("ERROR: Can't connect to remote FTP server %s", $ftpuri));
+                        next;
+                    };
+                    my $ftplist = ftpgetlist($ftp_node->[0], qr/^[^.]/);
+                    $tryftp = _search_file($ftplist, $find_file);
+                    if ($tryftp) { #  ! ! 
+                        my $file_dst = catfile($target_dir,$tryftp);
+                        $c->log_debug($pfx x 2, sprintf("Restore file %s from the FTP storage %s", $tryftp, $ftpuri));
+                        
+                        #  
+                        my $ftp_status = $c->fetch(
+                            -connect  => $ftpct,
+                            -protocol => 'ftp',
+                            -dir      => $target_dir,
+                            -cmd      => 'copy',
+                            -mode     => 'bin',
+                            -list     => $tryftp,
+                        );
+                        unless ($ftp_status && -e $file_dst) {
+                            $c->log_error($pfx, sprintf(" - [ERROR] Can't fetch file %s from %s to %s",$tryftp, $ftpuri, $target_dir));
+                            next
+                        }
+                        
+                        #       : 
+                        $meta = $self->_info($colls, host => $hostname, file => $tryftp) if value($ftpct => 'fixup');
+                        if ($meta) {
+                            my $size = -s $file_dst || 0;
+                            if (value($meta => "file_size") && value($meta => "file_size") ne $size){
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. Size got: %s; expected: %s",$file_dst, $size, uv2null(value($meta => "file_size"))));
+                                next;
+                            }
+                            my $md5 = md5sum($file_dst);
+                            if (value($meta => "file_md5") && value($meta => "file_md5") ne $md5) {
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. MD5 got: %s; expected: %s",$file_dst, $md5, uv2null(value($meta => "file_md5"))));
+                                next;
+                            }
+                            my $sha1 = sha1sum($file_dst);
+                            if (value($meta => "file_sha1") && value($meta => "file_sha1") ne $sha1) {
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. SHA1 got: %s; expected: %s",$file_dst, $sha1, uv2null(value($meta => "file_sha1"))));
+                                next;
+                            }
+                        }
+                        
+                        # OK!   
+                        $c->log_debug($pfx, sprintf(" - [ OK ] %s from %s",$tryftp, $ftpuri));
+                        $tbl->row(localtime2date_time, sprintf("%s File from FTP", $step), sprintf("%s from\n%s", $tryftp, $ftpuri), 'OK');
+                        $downloaded = 1;
+                        last;
+                    } else {
+                        $c->log_debug($pfx, sprintf(" - [SKIP] %s", $ftpuri));
+                    }
+                }
+                $c->log_debug($pfx x 2, "SKIPPED") unless $downloaded;
+                $tbl->row(localtime2date_time, sprintf("%s File from FTP", $step), '--- NONE ---' , 'SKIPPED') unless $downloaded;
+            }
+
+
+            # Step 02c.       HTTP     
+            $step = "Step 02c."; $c->log_debug($pfx, $step, "Search file on the HTTP storages");
+            if ($downloaded) {
+                $c->log_debug($pfx x 2, "SKIPPED");
+                $tbl->row(localtime2date_time, sprintf("%s File from HTTP", $step), '--- NONE ---' , 'SKIPPED');
+            } else {
+                #     HTTP 
+                my $http_node = _node2anode(node($job, $hostname => 'http'));
+                my $tryhttp;
+                foreach my $httpct (@$http_node) {
+                    my $http_uri = value($httpct, 'uri') || '';
+                    my $agent = new App::MBUtiny::CollectorAgent(
+                            uri         => $http_uri,
+                            user        => value($httpct, 'user'),
+                            password    => value($httpct, 'password'),
+                            timeout     => value($httpct, 'timeout'),
+                        );
+                    my $http_status = $agent->list(
+                            host        => $hostname,
+                        );
+                    #  
+                    if ($http_status) {
+                        my $httplist = array($agent->response => 'data/list');
+                        $tryhttp = _search_file($httplist, $find_file);
+                        #::debug(Dumper($httplist));
+                    } else {
+                        $c->log_error($pfx x 2, sprintf("ERROR %s: %s", $http_uri, unidecode($agent->error)));
+                        next;
+                    }
+                    #   
+                    if ($tryhttp) { #  ! ! 
+                        my $file_dst = catfile($target_dir,$tryhttp);
+                        $c->log_debug($pfx x 2, sprintf("Restore file %s from the HTTP storage %s", $tryhttp, $http_uri));
+                        
+                        #  
+                        my $download_status = $agent->download(
+                                host        => $hostname,
+                                file        => $tryhttp,
+                                path        => $file_dst,
+                            );
+                        if ($download_status) {
+                            $c->log_debug($pfx x 2, sprintf("OK %s: %s", $http_uri, unidecode(value($agent->response => 'data/message') || '')));
+                        } else {
+                            $c->log_error($pfx x 2, sprintf("ERROR %s: %s", $http_uri, unidecode($agent->error)));
+                            next;
+                        }
+                        unless (-e $file_dst) {
+                            $c->log_error($pfx, sprintf(" - [ERROR] Can't fetch file %s from %s to %s",$tryhttp, $http_uri, $target_dir));
+                            next
+                        }
+
+                        #       : 
+                        $meta = $self->_info($colls, host => $hostname, file => $tryhttp) if value($httpct => 'fixup');
+                        if ($meta) {
+                            my $size = -s $file_dst || 0;
+                            if (value($meta => "file_size") && value($meta => "file_size") ne $size){
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. Size got: %s; expected: %s",$file_dst, $size, uv2null(value($meta => "file_size"))));
+                                next;
+                            }
+                            my $md5 = md5sum($file_dst);
+                            if (value($meta => "file_md5") && value($meta => "file_md5") ne $md5) {
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. MD5 got: %s; expected: %s",$file_dst, $md5, uv2null(value($meta => "file_md5"))));
+                                next;
+                            }
+                            my $sha1 = sha1sum($file_dst);
+                            if (value($meta => "file_sha1") && value($meta => "file_sha1") ne $sha1) {
+                                $c->log_error($pfx, sprintf(" - [ERROR] File incorrect %s. SHA1 got: %s; expected: %s",$file_dst, $sha1, uv2null(value($meta => "file_sha1"))));
+                                next;
+                            }
+                        }
+
+                        # OK!   
+                        $c->log_debug($pfx, sprintf(" - [ OK ] %s from %s",$tryhttp, $http_uri));
+                        $tbl->row(localtime2date_time, sprintf("%s File from HTTP", $step), sprintf("%s from\n%s", $tryhttp, $http_uri), 'OK');
+                        $downloaded = 1;
+                        last;
+                    } else {
+                        $c->log_debug($pfx, sprintf(" - [SKIP] %s", $http_uri));
+                    }
+                }
+                $c->log_debug($pfx x 2, "SKIPPED") unless $downloaded;
+                $tbl->row(localtime2date_time, sprintf("%s File from HTTP", $step), '--- NONE ---' , 'SKIPPED') unless $downloaded;
+            }
+            
+            # Step 03.     (DATADIR)
+            $step = "Step 03."; $c->log_debug($pfx, $step, "Extraction");
+            if ($downloaded) {
+                my $x_ext = value($arcdef, 'arc'=>$arcname=>'ext') || 'zip';
+                $c->fextract(
+                    -in     => $target_dir,
+                    -out    => $target_dir,
+                    -method => 'ext',
+                    -list   => qr/\.$x_ext$/,
+                    -arcdef => $arcdef,
+                );
+                $c->log_debug($pfx x 2, 'OK');
+                $tbl->row(localtime2date_time, sprintf("%s Extraction", $step), $target_dir, 'OK');
+                my $oldshow = $self->show;
+                $self->show( ($oldshow ? ($oldshow."\n") : ""). sprintf("Your restored files located in \"%s\" directory for host %s", $target_dir, $hostname));
+                $status = 1;
+            } else {
+                $c->log_debug($pfx x 2, 'SKIPPED');
+                $tbl->row(localtime2date_time, sprintf("%s Extraction", $step), $target_dir, 'SKIPPED');
+            }
+            
+            #     
+            $tbl->hr;
+            $tbl->row(localtime2date_time, 'STATUS', '', $status ? 'OK' : 'ERROR');
+            my $tbl_rslt = $tbl->draw() || '';
+            $ret .= sprintf("Host: %s; Status: %s\n", $hostname, $status ? 'OK' : 'ERROR');
+            $ret .= sprintf("%s\n", $tbl_rslt);
+            
+            $c->log_debug(sprintf("--> Done: \"%s\". %s", $hostname, $status ? 'OK' : 'ERROR'));
+        } else {
+            $ret .= sprintf("Host: %s; Status: %s\n\n", $hostname, 'SKIP');
+            $c->log_debug(sprintf("--> Skipped \"%s\". Enable flag is off", $hostname));
+        }
+
+    }
+    $c->log_debug("Finish processing hosts");    
+    
+    $self->msg($ret);
+    
+    return 1;
+}
 
 sub get_jobs { #   .             
     my $self = shift;
@@ -1185,6 +1571,25 @@ sub _del { #
     
     return $gst;
 }
+sub _info { #        
+    my $self = shift;
+    my $colls = shift; #       "
+    croak("Second argument must be reference to array!") unless $colls && ref($colls) eq 'ARRAY';
+    my %data = @_;
+    my $c = $self->c;
+    
+    foreach my $cl (@$colls) {
+        if ($cl->info(%data)) {
+            $c->log_debug(" " x 6, sprintf("COLLECTOR> INFO OK %s: %s", $cl->{uri}, unidecode(value($cl->response => 'data/message') || '')));
+            my $reth = hash($cl->response => 'data');
+            return {%$reth};
+        } else {
+            $c->log_error(" " x 6, sprintf("COLLECTOR> INFO ERROR %s: %s", $cl->{uri}, unidecode($cl->error) || ''));
+        }
+    }
+    
+    return undef;
+}
 
 sub DESTROY {
     my $self = shift;
@@ -1255,6 +1660,24 @@ sub _node_correct { #
     }
     return [@nc];
 }
+sub _search_file { #       
+    my $list = shift;
+    my $file = shift;
+    return undef unless $list && ref($list) eq 'ARRAY';
+    
+    my $s = undef;
+    my $l = undef;
+    foreach (sort {$a cmp $b} @$list) {
+        $s = $_;
+        return $s if $file && $s eq $file;
+        if ($file && $s gt $file) {
+            return $l;
+        }
+        $l = $s;
+        
+    }
+    return $s
+}
 
 1;
 
